package cellar

import (
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"os/signal"
	"strconv"
	"strings"
	"time"

	coreiface "github.com/ipfs/interface-go-ipfs-core"

	"github.com/marc-gr/cellar/config"
)

type Cellar struct {
	aesKey              []byte
	ipfsNodeAddr        string
	publishKey          coreiface.Key
	latestVersionSynced int
}

func Run() error {
	if err := os.Setenv("LIBP2P_FORCE_PNET", "1"); err != nil {
		return fmt.Errorf("setenv: %w", err)
	}

	if _, err := os.Stat(config.CellarDir() + "/pid"); err == nil {
		return errors.New("cellar already running")
	}

	pidFile, err := os.Create(config.CellarDir() + "/pid")
	if err != nil {
		return fmt.Errorf("create pid file: %w", err)
	}

	if _, err := pidFile.WriteString(fmt.Sprint(os.Getpid())); err != nil {
		_ = pidFile.Close()
		return fmt.Errorf("write pid file: %w", err)
	}

	if err := pidFile.Close(); err != nil {
		return fmt.Errorf("close pid file: %w", err)
	}
	defer func() { _ = os.Remove(config.CellarDir() + "/pid") }()

	cmdIPFS := exec.Command(
		"ipfs",
		"-c", config.IPFSPath(),
		"daemon",
	)
	cmdIPFS.Stdout = os.Stdout
	cmdIPFS.Stderr = os.Stderr

	if err := cmdIPFS.Start(); err != nil {
		return fmt.Errorf("start ipfs: %w", err)
	}

	cmdIPFSCluster := exec.Command(
		"ipfs-cluster-service",
		"-c", config.IPFSClusterPath(),
		"daemon",
	)
	cmdIPFSCluster.Stdout = os.Stdout
	cmdIPFSCluster.Stderr = os.Stderr

	if err := cmdIPFSCluster.Start(); err != nil {
		_ = cmdIPFS.Process.Signal(os.Interrupt)
		return fmt.Errorf("start cluster: %w", err)
	}

	time.Sleep(5 * time.Second)

	self, err := resolveSelf()
	if err != nil {
		return fmt.Errorf("get self address: %w", err)
	}

	aesKey, err := config.LoadSecretAESKey()
	if err != nil {
		return err
	}

	key, err := config.GeneratePublishKey(aesKey)
	if err != nil {
		return err
	}

	c := Cellar{
		aesKey:       aesKey,
		ipfsNodeAddr: self,
		publishKey:   key,
	}

	if err := c.discoverLoop(); err != nil {
		return fmt.Errorf("discover: %w", err)
	}

	go c.syncloop()

	sig := make(chan os.Signal, 1)
	signal.Notify(sig, os.Interrupt)

	interrupt := <-sig

	close(sig)

	_ = cmdIPFSCluster.Process.Signal(interrupt)
	_ = cmdIPFSCluster.Wait()

	_ = cmdIPFS.Process.Signal(interrupt)
	_ = cmdIPFS.Wait()

	return nil
}

func Shutdown() error {
	all, err := ioutil.ReadFile(config.CellarDir() + "/pid")
	if err != nil {
		return err
	}
	pid, err := strconv.ParseInt(string(all), 10, 64)
	if err != nil {
		return err
	}
	proc, err := os.FindProcess(int(pid))
	if err != nil {
		return err
	}
	return proc.Signal(os.Interrupt)
}

func resolveSelf() (string, error) {
	cmdID := exec.Command(
		"ipfs",
		"-c", config.IPFSPath(),
		"id",
	)
	out, err := cmdID.CombinedOutput()
	if err != nil {
		return "", err
	}
	addrs := struct {
		Addresses []string
	}{}
	if err := json.Unmarshal(out, &addrs); err != nil {
		return "", err
	}
	if len(addrs.Addresses) == 0 {
		return "", errors.New("can't get self address")
	}
	addr := func() string {
		for _, addr := range addrs.Addresses {
			if strings.HasPrefix(addr, "/ip4/127.0.0.1") ||
				strings.HasPrefix(addr, "/ip6/::1") {
				continue
			}
			return addr
		}
		return ""
	}()
	if addr == "" {
		return "", errors.New("not available valid self address")
	}
	return addr, nil
}
