package config

import (
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"os/user"
)

var aesKeypath = CellarDir() + "/keystore"

const (
	aesKeyname    = "id.key"
	clusterSecret = "b82c5ec5b778f1e695907b5ba3949cdbd325d35589a196ea78b67ac1fcf22b08"

	cellarDir      = "/.cellar"
	ipfsDir        = "/ipfs"
	ipfsClusterDir = "/cluster"
)

func InitConfig() error {
	hasConfig, err := alreadyHasConfig()
	if err != nil {
		return err
	}

	if hasConfig {
		return nil
	}

	if err := initIPFS(); err != nil {
		return err
	}

	if err := initIPFSCluster(); err != nil {
		return err
	}

	mkdirCmd := exec.Command("mkdir", "-p", aesKeypath)
	if out, err := mkdirCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	aesKey, err := NewSecretKey()
	if err != nil {
		return err
	}

	keyFilepath := fmt.Sprintf(
		"%s/%s",
		aesKeypath,
		aesKeyname,
	)

	f, err := os.Create(keyFilepath)
	if err != nil {
		return err
	}
	defer f.Close()

	if _, err := f.Write(aesKey); err != nil {
		return err
	}

	fmt.Println(`
	You created a new cellar config. Your ID file is at`, keyFilepath, `
	Please store this file cautiously, as if you start cellar in any other machine you will need it to access your files.
	`)

	return nil
}

func initIPFS() error {
	mkdirCmd := exec.Command("mkdir", "-p", IPFSPath())
	if out, err := mkdirCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	initCmd := exec.Command(
		"ipfs",
		"-c", IPFSPath(),
		"init",
	)
	if out, err := initCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	removeBootstrapCmd := exec.Command(
		"ipfs",
		"-c", IPFSPath(),
		"bootstrap", "rm", "--all",
	)
	if out, err := removeBootstrapCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	if err := SaveSwarmKey(); err != nil {
		return err
	}

	return nil
}

func initIPFSCluster() error {
	if err := os.Setenv("CLUSTER_SECRET", clusterSecret); err != nil {
		return err
	}

	mkdirCmd := exec.Command("mkdir", "-p", IPFSClusterPath())
	if out, err := mkdirCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	initCmd := exec.Command(
		"ipfs-cluster-service",
		"-c", IPFSClusterPath(),
		"init",
	)
	if out, err := initCmd.CombinedOutput(); err != nil {
		return errors.New(string(out))
	}

	return nil
}

func alreadyHasConfig() (bool, error) {
	{
		info, err := os.Stat(CellarDir())
		if err != nil {
			// only fails if does not exist
			return false, nil
		}
		if !info.IsDir() {
			return false, errors.New(CellarDir() + " not a directory")
		}
	}

	{
		info, err := os.Stat(aesKeypath)
		if err != nil {
			// only fails if does not exist
			return false, nil
		}
		if !info.IsDir() {
			return false, errors.New(aesKeypath + " not a directory")
		}
	}

	{
		keyFilepath := fmt.Sprintf(
			"%s/%s",
			aesKeypath,
			aesKeyname,
		)
		info, err := os.Stat(keyFilepath)
		if err != nil {
			// only fails if does not exist
			return false, nil
		}
		if info.IsDir() {
			return false, errors.New(keyFilepath + " is a directory")
		}
	}

	return true, nil
}

func CellarDir() string {
	usr, err := user.Current()
	if err != nil {
		panic(err)
	}
	return usr.HomeDir + cellarDir
}

func IPFSPath() string {
	return CellarDir() + ipfsDir
}

func IPFSClusterPath() string {
	return CellarDir() + ipfsClusterDir
}

func LoadSecretAESKey() ([]byte, error) {
	keyFilepath := fmt.Sprintf(
		"%s/%s",
		aesKeypath,
		aesKeyname,
	)
	return ioutil.ReadFile(keyFilepath)
}
