package config

import (
	"bytes"
	"errors"
	"os/exec"

	"github.com/ipfs/go-ipfs/keystore"
	ipfspath "github.com/ipfs/go-path"
	coreiface "github.com/ipfs/interface-go-ipfs-core"
	path "github.com/ipfs/interface-go-ipfs-core/path"
	crypto "github.com/libp2p/go-libp2p-core/crypto"
	peer "github.com/libp2p/go-libp2p-core/peer"
)

var keystoreDir = IPFSPath() + "/keystore"

const KeyName = "publishkey"

type key struct {
	name   string
	peerID peer.ID
}

// Name returns the key name
func (k key) Name() string {
	return k.name
}

// Path returns the path of the key.
func (k key) Path() path.Path {
	return path.New(ipfspath.Join([]string{"/ipns", k.peerID.Pretty()}))
}

// ID returns key PeerID
func (k key) ID() peer.ID {
	return k.peerID
}

// GeneratePublishKey generates new key, stores it in the keystore under the specified
// name and returns a base58 encoded multihash of its public key.
func GeneratePublishKey(seed []byte) (coreiface.Key, error) {
	mkdirCmd := exec.Command("mkdir", "-p", keystoreDir)
	if out, err := mkdirCmd.CombinedOutput(); err != nil {
		return nil, errors.New(string(out))
	}

	ks, err := keystore.NewFSKeystore(keystoreDir)
	if err != nil {
		return nil, err
	}

	priv, pub, err := crypto.GenerateEd25519Key(bytes.NewReader(seed))
	if err != nil {
		return nil, err
	}

	_ = ks.Delete(KeyName)

	if err := ks.Put(KeyName, priv); err != nil {
		return nil, err
	}

	pid, err := peer.IDFromPublicKey(pub)
	if err != nil {
		return nil, err
	}

	return key{KeyName, pid}, nil
}
