package storage

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"strings"

	"github.com/marc-gr/cellar/config"
	"github.com/marc-gr/cellar/crypto"
)

func AddFilepaths(filePaths ...string) ([]byte, error) {
	var (
		combinedOut []byte
	)

	for _, filePath := range filePaths {
		ipfsFilepaths, cids, addOut, err := Add("/", filePath)
		combinedOut = append(combinedOut, addOut...)
		if err != nil {
			return combinedOut, err
		}

		for i, ipfsFilepath := range ipfsFilepaths {
			pinOut, err := ClusterPin(ipfsFilepath, cids[i])
			combinedOut = append(combinedOut, pinOut...)
			if err != nil {
				return combinedOut, err
			}
		}
	}

	return combinedOut, nil
}

func Add(base, filePath string) ([]string, []string, []byte, error) {
	var ipfsFilepaths []string
	var cids []string
	info, err := os.Stat(filePath)
	if err != nil {
		return nil, nil, nil, err
	}

	wasDir := info.IsDir()

	originalBase := base
	base += filepath.Base(filePath)

	var retOut []byte
	err = filepath.Walk(filePath,
		func(path string, info os.FileInfo, err error) error {
			if err != nil {
				return fmt.Errorf("walk: %w", err)
			}

			if wasDir && path == filePath {
				return nil
			}

			if wasDir && filepath.Base(base) !=
				filepath.Base(filepath.Dir(path)) {
				return nil
			}

			if info.IsDir() {
				ipaths, icids, out, err := Add(base+"/", path)
				if err != nil {
					retOut = out
					return fmt.Errorf("add dir: %w", err)
				}
				ipfsFilepaths = append(ipfsFilepaths, ipaths...)
				cids = append(cids, icids...)
				return nil
			}

			var ipfsFilepath string
			if wasDir {
				ipfsFilepath = fmt.Sprintf("%s/%s",
					base,
					filepath.Base(path),
				)
			} else {
				ipfsFilepath = base
			}
			ipfsFilepaths = append(ipfsFilepaths, ipfsFilepath)

			cipheredPath, err := cipherFile(path)
			if err != nil {
				return err
			}
			defer os.Remove(cipheredPath)

			writeCmd := exec.Command(
				"ipfs",
				"-c", config.IPFSPath(),
				"add", cipheredPath,
			)

			fmt.Println(
				"ipfs",
				"-c", config.IPFSPath(),
				"add", cipheredPath,
			)

			out, err := writeCmd.CombinedOutput()
			if err != nil {
				retOut = out
				return fmt.Errorf("create file: %w", err)
			}

			scanner := bufio.NewScanner(bytes.NewReader(out))
			scanner.Scan()
			line := strings.Split(scanner.Text(), " ")
			var cid string
			for i, l := range line {
				if l == filepath.Base(cipheredPath) {
					cid = line[i-1]
				}
			}
			if cid == "" {
				retOut = out
				return errors.New("unexpected response")
			}

			cids = append(cids, cid)
			if originalBase == "" {
				return nil
			}
			return AddFileToStore(ipfsFilepath, cid)
		},
	)

	return ipfsFilepaths, cids, retOut, err
}

func cipherFile(path string) (string, error) {
	key, err := config.LoadSecretAESKey()
	if err != nil {
		return "", err
	}

	data, err := ioutil.ReadFile(path)
	if err != nil {
		return "", err
	}

	ciphered, err := crypto.Cipher(data, key)
	if err != nil {
		return "", err
	}

	cipheredPath := path + ".ciph"
	// create a new file for saving the encrypted data.
	f, err := os.Create(cipheredPath)
	if err != nil {
		return "", err
	}
	defer f.Close()

	_, err = io.Copy(f, bytes.NewReader(ciphered))
	if err != nil {
		return "", err
	}

	return cipheredPath, nil
}

func ClusterPin(filepath, cid string) ([]byte, error) {
	pinCmd := exec.Command(
		"ipfs-cluster-ctl",
		"pin", "add",
		cid,
	)

	fmt.Println(
		"ipfs-cluster-ctl",
		"pin", "add",
		cid,
	)

	out, err := pinCmd.CombinedOutput()
	if err != nil {
		return out, fmt.Errorf("pin add: %w", err)
	}
	return out, nil
}
