package storage

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"strconv"
	"strings"

	"github.com/marc-gr/cellar/config"
	"github.com/marc-gr/cellar/crypto"
)

var StoreFile = config.CellarDir() + "/store.list"

func AddFileToStore(filename, cid string) error {
	version, files, err := ReadStoreFile()
	if err != nil {
		return err
	}
	files[filename] = cid
	version++
	return SaveStoreFile(version, files)
}

func SaveStoreFile(version int, files map[string]string) error {
	key, err := config.LoadSecretAESKey()
	if err != nil {
		return err
	}

	data := fmt.Sprintln(version)
	for f, cid := range files {
		data += fmt.Sprintln(f + "," + cid)
	}

	storeFile, err := os.Create(
		StoreFile,
	)
	if err != nil {
		return err
	}
	defer storeFile.Close()
	if _, err := storeFile.Seek(0, 0); err != nil {
		return err
	}

	ciphered, err := crypto.Cipher([]byte(data), key)
	if err != nil {
		return err
	}
	if _, err := storeFile.Write(ciphered); err != nil {
		return err
	}
	return nil
}

func ReadStoreFile() (int, map[string]string, error) {
	key, err := config.LoadSecretAESKey()
	if err != nil {
		return 0, nil, err
	}

	storeFile, err := os.OpenFile(
		StoreFile,
		os.O_CREATE|os.O_RDWR,
		0644,
	)
	if err != nil {
		return 0, nil, err
	}
	defer storeFile.Close()

	contents, err := ioutil.ReadAll(storeFile)
	if err != nil {
		return 0, nil, err
	}

	var data []byte
	if len(contents) > 0 {
		var err error
		data, err = crypto.Decipher(contents, key)
		if err != nil {
			return 0, nil, err
		}
	}
	scanner := bufio.NewScanner(bytes.NewReader(data))
	scanner.Scan()
	version := func() int {
		if scanner.Text() == "" {
			return 0
		}
		version, err := strconv.ParseInt(scanner.Text(), 10, 64)
		if err != nil {
			return 0
		}
		return int(version)
	}()

	files := map[string]string{}
	for scanner.Scan() {
		if scanner.Err() != nil {
			return 0, nil, scanner.Err()
		}
		line := strings.Split(scanner.Text(), ",")
		if len(line) != 2 {
			return 0, nil, errors.New("malformed store line")
		}
		files[line[0]] = line[1]
	}

	return version, files, nil
}
