package cellar

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"log"
	"os"
	"os/exec"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/marc-gr/cellar/config"
	"github.com/marc-gr/cellar/crypto"
	"github.com/marc-gr/cellar/storage"
)

var storeDir = config.CellarDir() + "/files"

func (c *Cellar) syncloop() {
	if err := c.initStorelist(); err != nil {
		log.Println("sync init store:", err)
		time.Sleep(5 * time.Second)
		go c.syncloop()
		return
	}
	ticker := time.NewTicker(10 * time.Second) // for testing purposes, should be higher
	mkdirCmd := exec.Command("mkdir", "-p", storeDir)
	if out, err := mkdirCmd.CombinedOutput(); err != nil {
		log.Println(string(out))
		log.Println("sync mkdir:", err)
		return
	}

loop:
	for range ticker.C {
		version, files, err := storage.ReadStoreFile()
		if err != nil {
			log.Println("sync loop:", err)
			continue
		}
		if version <= c.latestVersionSynced {
			log.Println("nothing to sync, skipping")
			continue
		}
		log.Println("updating from", c.latestVersionSynced, "to", version)
		for file, cid := range files {
			if err := c.syncfile(file, cid); err != nil {
				log.Println("sync file:", err)
				continue loop
			}
		}
		if err := c.publishStorelist(); err != nil {
			log.Println("sync:", err)
			continue
		}
		c.latestVersionSynced = version
	}
}

func (c *Cellar) publishStorelist() error {
	paths, cids, _, err := storage.Add("", storage.StoreFile)
	if err != nil {
		return err
	}
	if out, err := storage.ClusterPin(paths[0], cids[0]); err != nil {
		return fmt.Errorf("%s: %w", string(out), err)
	}
	publishCmd := exec.Command(
		"ipfs",
		"-c", config.IPFSPath(),
		"name", "publish", "--key="+c.publishKey.Name(),
		"/ipfs/"+cids[0],
	)
	out, err := publishCmd.CombinedOutput()
	log.Println(string(out))
	if err != nil {
		return err
	}
	return nil
}

func (c *Cellar) syncfileto(file, cid, to string) error {
	log.Println("syncing", file, ":", cid)
	mkdirCmd := exec.Command("mkdir", "-p", to)
	if _, err := mkdirCmd.CombinedOutput(); err != nil {
		return err
	}

	data, err := c.getRemoteData(cid)
	if err != nil {
		return err
	}

	f, err := os.Create(storeDir + file)
	if err != nil {
		return err
	}
	defer f.Close()
	if _, err := f.Write(data); err != nil {
		return err
	}
	return nil
}

func (c *Cellar) getRemoteData(cid string) ([]byte, error) {
	catCmd := exec.Command(
		"ipfs",
		"-c", config.IPFSPath(),
		"cat", cid,
	)

	fmt.Println(
		"ipfs",
		"-c", config.IPFSPath(),
		"cat", cid,
	)

	out, err := catCmd.CombinedOutput()
	if err != nil {
		return nil, fmt.Errorf("cat: %v: %w", string(out), err)
	}

	return crypto.Decipher(out, c.aesKey)
}

func (c *Cellar) syncfile(file, cid string) error {
	return c.syncfileto(file, cid, filepath.Dir(storeDir+file))
}

func (c *Cellar) loadRemoteStoreFile() (int, map[string]string, error) {
	resolveCmd := exec.Command(
		"ipfs",
		"-c", config.IPFSPath(),
		"name", "resolve",
		c.publishKey.ID().String(),
	)

	fmt.Println(
		"ipfs",
		"-c", config.IPFSPath(),
		"name", "resolve",
		c.publishKey.ID().String(),
	)

	out, err := resolveCmd.CombinedOutput()
	if err != nil || len(out) == 0 {
		log.Println("sync: remote store list not found")
		log.Println("sync:", string(out))
		return 0, map[string]string{}, nil
	}

	ciphStoreFile, err := c.getRemoteData(string(out[:52]))
	if err != nil {
		return 0, nil, fmt.Errorf("get remote data: %v: %w", string(out), err)
	}
	data, err := crypto.Decipher(ciphStoreFile, c.aesKey)
	if err != nil {
		return 0, nil, fmt.Errorf("decipher store file: %v: %w", string(out), err)
	}
	fmt.Println("sync: remote store.list contents:", string(data))

	scanner := bufio.NewScanner(bytes.NewReader(data))
	scanner.Scan()
	version := func() int {
		if scanner.Text() == "" {
			return 0
		}
		version, err := strconv.ParseInt(scanner.Text(), 10, 64)
		if err != nil {
			return 0
		}
		return int(version)
	}()

	files := map[string]string{}
	for scanner.Scan() {
		if scanner.Err() != nil {
			return 0, nil, scanner.Err()
		}
		line := strings.Split(scanner.Text(), ",")
		if len(line) != 2 {
			return 0, nil, errors.New("malformed store line")
		}
		files[line[0]] = line[1]
	}

	return version, files, nil
}

func (c *Cellar) initStorelist() error {
	remoteVersion, remoteFiles, err := c.loadRemoteStoreFile()
	if err != nil {
		return fmt.Errorf("loadRemote: %w", err)
	}
	version, files, err := storage.ReadStoreFile()
	if err != nil {
		return fmt.Errorf("read store: %w", err)
	}
	switch {
	case version > remoteVersion:
		if err := c.publishStorelist(); err != nil {
			return fmt.Errorf("version > remoteVersion: %w", err)
		}
	case version < remoteVersion:
		for rf, rcid := range remoteFiles {
			files[rf] = rcid
		}
		if err := storage.SaveStoreFile(remoteVersion+1, files); err != nil {
			return fmt.Errorf("save store: %w", err)
		}
		if err := c.publishStorelist(); err != nil {
			return fmt.Errorf("publish: %w", err)
		}
	}
	return nil
}
